//*****************************************************************************
//  Copyright (c) 2021 Glenfly Tech Co., Ltd.. 
//  All Rights Reserved.
//
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Glenfly Tech Co., Ltd..;
//  the contents of this file may not be disclosed to third parties, copied or
//  duplicated in any form, in whole or in part, without the prior written
//  permission of Glenfly Tech Co., Ltd..
//
//  The copyright of the source code is protected by the copyright laws of the People's
//  Republic of China and the related laws promulgated by the People's Republic of China
//  and the international covenant(s) ratified by the People's Republic of China. 
//*****************************************************************************

#include <linux/proc_fs.h>
#include <linux/slab.h>
#include <linux/videodev2.h>

#include "gf.h"
#include "gf_capture.h"
#include "cam_mgr.h"

#ifndef VIP_MGR_VERSION_CODE
#define VIP_MGR_VERSION_CODE LINUX_VERSION_CODE
#endif


//in log macro, dev means cam_mgr pointer
#define mgr_printk(level, dev, fmt, arg...)       \
    printk(level "[%s] %s: " fmt "\n", (dev)->name, GF_FUNC_NAME(__func__), ## arg)

#define mgr_err(dev, fmt, arg...)                 \
    mgr_printk(KERN_ERR, dev, fmt, ## arg)

#define mgr_warn(dev, fmt, arg...)                \
    mgr_printk(KERN_WARNING, dev, fmt, ## arg)

#define mgr_info(dev, fmt, arg...)                \
    mgr_printk(KERN_INFO, dev, fmt, ## arg)

#define mgr_dbg(dev, fmt, arg...)                 \
    mgr_printk(KERN_DEBUG, dev, fmt, ## arg)

#define CAM_MGR_DEBUG
#ifdef CAM_MGR_DEBUG

enum LOG_LEVEL {
    LOG_ERR = 0,
    LOG_WARNING,
    LOG_INFO,
    LOG_DEBUG,
};

static int mgr_debuglevel = LOG_WARNING;

#define dprintk(loglevel, dev, fmt, arg...)                 \
    do {                                                    \
        if(loglevel <= (mgr_debuglevel))                    \
            mgr_printk(KERN_WARNING, dev, fmt, ## arg);     \
    }while(0)

#else

#define dprintk(loglevel, dev, fmt, arg...)

#endif

#define UNUSED(i)               ((void)i)

#define REFRESHRATE_DEN         (100)

#define HARDWARE_BUFFER_MAX         3
#define VIP_ENABLE_BUFFER_NUM       3
#define WB_ENABLE_BUFFER_NUM        2
#define VIP_SLOT_BUFFER_NUM         3
#define WB_SLOT_BUFFER_NUM          1

typedef void (*capture_cb)(gf_capture_id_t id, void *data, unsigned int flags);

struct cam_ops {
    int (*init)(struct cam_mgr *pmgr);
    int (*deinit)(struct cam_mgr *mgr);

    int (*get_caps_num)(struct cam_mgr *mgr);
    int (*get_caps)(struct cam_mgr *mgr, struct cam_caps *caps);
    int (*set_format)(struct cam_mgr *mgr, struct cam_fmt *fmt);

    int (*set_callback)(struct cam_mgr *mgr, capture_cb cb);

    int (*enable)(struct cam_mgr *mgr);
    int (*disable)(struct cam_mgr *mgr);

    int (*set_buffer)(struct cam_mgr *mgr, unsigned int slot, unsigned long long addr);
};

enum {
    CAM_VIP     = 0,
    CAM_WB      = 1,
};

/**
 *  state-change:
 *
 *  INIT ---> DEINIT
 *  INIT ---> RUNNING
 *  RUNNING <---> PAUSED
 *  RUNNING ---> STOPPED
 *  PAUSED ---> STOPPED
 *  STOPPED ---> DEINIT
 */
enum {
    CAM_STATE_NONE = 0,
    CAM_STATE_INIT,
    CAM_STATE_RUNNING,
    CAM_STATE_PAUSED,
    CAM_STATE_STOPPED,
    CAM_STATE_DEINIT,
};

struct cam_state {
    spinlock_t lock;
    int cur;
    int pending;
};

struct cam_mgr {
    struct proc_dir_entry * proc_entry;
    char            name[16];
    unsigned int    type;
    gf_capture_id_t id;
    struct cam_ops *cam_ops;

    struct list_head    queue;
    spinlock_t          queue_lock;
    unsigned int        queue_cnt;

    unsigned int        enable_cnt;
    struct cam_buf *    hw_buf[HARDWARE_BUFFER_MAX];
    unsigned int        buf_get;
    unsigned int        buf_set;
    unsigned int        slot_cnt;
    unsigned int        cur_slot;

    struct cam_fmt fmt;
    struct cam_state state;

    //******************************
    // V4L2 driver cb
    update_buffer_func  cb;
    void                *cb_data;

    // capture cb
    PFN_GF_CAPTURE_OPS_T ops;
    void *disp;

    //******************************
    // below for debug use
    int64_t irq_count;
    int64_t notify_count;
    struct {
        unsigned notify_disable : 1;
    };
};

#define QUEUE_LOCK(mgr)     \
    spin_lock_irq(&mgr->queue_lock)

#define QUEUE_UNLOCK(mgr)   \
    spin_unlock_irq(&mgr->queue_lock)


#define STATE_LOCK(mgr)     \
    spin_lock_irq(&mgr->state.lock);

#define STATE_UNLOCK(mgr)   \
    spin_unlock_irq(&mgr->state.lock);

static inline int STATE_CURRENT(struct cam_mgr *mgr) {
    struct cam_state *state = &mgr->state;

    return state->cur;
}

static inline void STATE_INITIALIZE(struct cam_mgr *mgr) {
    struct cam_state *state = &mgr->state;

    spin_lock_init(&state->lock);
    state->cur = CAM_STATE_NONE;
    state->pending = CAM_STATE_NONE;
}

static inline void STATE_TRANSITION_SET(struct cam_mgr *mgr, int next) {
    struct cam_state *state = &mgr->state;

    state->cur = next;
    state->pending = CAM_STATE_NONE;
}
static inline void STATE_TRANSITION_START(struct cam_mgr *mgr, int next) {
    struct cam_state *state = &mgr->state;

    state->pending = next;
}

static inline void STATE_TRANSITION_CANCEL(struct cam_mgr *mgr) {
    struct cam_state *state = &mgr->state;

    state->pending = CAM_STATE_NONE;
}

static inline void STATE_TRANSITION_COMMIT(struct cam_mgr *mgr) {
    struct cam_state *state = &mgr->state;

    state->cur = state->pending;
    state->pending = CAM_STATE_NONE;
}

/********************************************************************/

static int cam_buf_is_singaled(struct cam_buf *buf) {
    if (buf->ops && buf->ops->is_signaled)
        return buf->ops->is_signaled(buf);
    else    //default ingore fence
        return 1;
}

static unsigned long long cam_buf_get_addr(struct cam_buf *buf) {
    if (buf->ops && buf->ops->get_addr)
        return buf->ops->get_addr(buf);
    else {
        return buf->handle;
    }
}

static void *cam_buf_get_cpu_vaddr(struct cam_buf *buf) {
    if (buf->ops && buf->ops->get_cpu_vaddr)
        return buf->ops->get_cpu_vaddr(buf);
    else {
        return NULL;
    }
}

static const char *cam_buf_flag_str[] = {
    "available",    //0 CAM_BUF_AVAILABLE
    "used",         //1 CAM_BUF_USED
    "busy",         //2 CAM_BUF_BUSY
    "unused",       //3 CAM_BUF_UNUSED
    "unknown"       //4 error flag
};

static inline const char *cam_buf_flag_2_str(struct cam_buf *buf) {
    if (buf->flag < CAM_BUF_AVAILABLE || buf->flag > CAM_BUF_UNUSED)
        return cam_buf_flag_str[4];

    return cam_buf_flag_str[buf->flag];
}

/********************************************************************/

static inline uint32_t v4l2_fmt_2_capture_fmt(uint32_t v4l2_fmt) {
    uint32_t capture_fmt;

    switch (v4l2_fmt) {
        case V4L2_PIX_FMT_RGB32:
            capture_fmt = GF_VIP_FMT_RGB444_24BIT_SDR;
            break;
        case V4L2_PIX_FMT_NV24:
            capture_fmt = GF_VIP_FMT_YCBCR444_24BIT_SDR;
            break;
        case V4L2_PIX_FMT_YUYV:
            capture_fmt = GF_VIP_FMT_YCBCR422_8BIT_SDR_ES;
            break;
        default:
            capture_fmt = GF_VIP_FMT_RGB444_24BIT_SDR;
    }

    return capture_fmt;
}

static inline uint32_t capture_fmt_2_v4l2_fmt(uint32_t capture_fmt) {
    uint32_t v4l2_format;

    switch (capture_fmt) {
        case GF_VIP_FMT_RGB444_24BIT_SDR:
            v4l2_format = V4L2_PIX_FMT_RGB32;
            break;
        case GF_VIP_FMT_YCBCR444_24BIT_SDR:
            v4l2_format = V4L2_PIX_FMT_NV24;
            break;
        case GF_VIP_FMT_YCBCR422_8BIT_SDR_ES:
        case GF_VIP_FMT_YCBCR422_8BIT_DDR_ES:
            v4l2_format = V4L2_PIX_FMT_YUYV;
            break;
        default:
            v4l2_format = V4L2_PIX_FMT_RGB32;
    }

    return v4l2_format;
}

/********************************************************************/
//
//  vip ops
//
/********************************************************************/

static int vip_init(struct cam_mgr *mgr) {
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_INIT;
    param.init_param.fbNum = mgr->slot_cnt;

    return mgr->ops(mgr->disp, id, &param);
}

static int vip_deinit(struct cam_mgr *mgr) {
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_DEINIT;

    return mgr->ops(mgr->disp, id, &param);
}

static int vip_enable(struct cam_mgr *mgr) {
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_ENABLE;

    return mgr->ops(mgr->disp, id, &param);
}

static int vip_disable(struct cam_mgr *mgr) {
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_DISABLE;

    return mgr->ops(mgr->disp, id, &param);
}

static int vip_get_caps_num(struct cam_mgr *mgr) {
    int ret = 0;
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};

    dprintk(LOG_DEBUG, mgr, "enter");
    // first time get modeNum
    param.op = GF_CAPTURE_OP_QUERY_CAPS;
    param.query_caps_param.modeNum = 0;
    param.query_caps_param.mode = NULL;
    ret = mgr->ops(mgr->disp, id, &param);
    if (ret || param.query_caps_param.modeNum == 0) {
        return 0;
    }

    return param.query_caps_param.modeNum;
}

static int vip_get_caps(struct cam_mgr *mgr, struct cam_caps *caps) {
    int ret = 0;
    int i;
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};
    gf_vip_mode_t *mode;

    dprintk(LOG_DEBUG, mgr, "enter");
    mode = kmalloc(sizeof(gf_vip_mode_t) * caps->caps_num, GFP_KERNEL | __GFP_ZERO);
    // second time get mode
    param.op = GF_CAPTURE_OP_QUERY_CAPS;
    param.query_caps_param.modeNum = (unsigned int) caps->caps_num;
    param.query_caps_param.mode = mode;
    ret = mgr->ops(mgr->disp, id, &param);
    if (!ret) {
        for (i = 0; i < param.query_caps_param.modeNum && i < caps->caps_num; i++) {
            caps->formats[i].width = mode[i].xRes;
            caps->formats[i].height = mode[i].yRes;
            caps->formats[i].format = capture_fmt_2_v4l2_fmt(mode[i].fmt);
            caps->formats[i].fps_num = mode[i].refreshrate;
            caps->formats[i].fps_den = REFRESHRATE_DEN;
            dprintk(LOG_WARNING, mgr, "query_caps %x %dx%d @ %d/%d fps",
                caps->formats[i].format, caps->formats[i].width, caps->formats[i].height, caps->formats[i].fps_num, caps->formats[i].fps_den);
        }
        if (param.query_caps_param.modeNum < caps->caps_num)
            caps->caps_num = param.query_caps_param.modeNum;
    }
    kfree(mode);

    return ret;
}

static int vip_set_format(struct cam_mgr *mgr, struct cam_fmt *fmt) {
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_SET_MODE;
    param.set_mode_param.fmt = v4l2_fmt_2_capture_fmt(fmt->format);
    param.set_mode_param.chip = GF_VIP_CARD_AVD7611;
    param.set_mode_param.xres = fmt->width;
    param.set_mode_param.yres = fmt->height;
    param.set_mode_param.refreshrate = fmt->fps_num * REFRESHRATE_DEN / fmt->fps_den;

    dprintk(LOG_WARNING, mgr, "set format %x %dx%d @ %d/%d fps", fmt->format, fmt->width, fmt->height, fmt->fps_num, fmt->fps_den);
    return mgr->ops(mgr->disp, id, &param);
}

static int vip_set_callback(struct cam_mgr *mgr, capture_cb cb) {
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_SET_CALLBACK;
    param.set_callback_param.cb = cb;
    param.set_callback_param.cb_data = mgr;

    return mgr->ops(mgr->disp, id, &param);
}

static int vip_set_buffer(struct cam_mgr *mgr, unsigned int slot, unsigned long long addr) {
    gf_capture_id_t id = mgr->id;
    vip_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_SET_BUFF;
    param.set_buffer_param.fbIdx = slot;
    param.set_buffer_param.fbAddr = addr;

    return mgr->ops(mgr->disp, id, &param);
}


static struct cam_ops vip_ops = {
    .init           = vip_init,
    .deinit         = vip_deinit,
    .enable         = vip_enable,
    .disable        = vip_disable,
    .get_caps_num   = vip_get_caps_num,
    .get_caps       = vip_get_caps,
    .set_format     = vip_set_format,
    .set_callback   = vip_set_callback,
    .set_buffer     = vip_set_buffer,
};

/************************************************************************/
//
//  write-back ops
//
/************************************************************************/

static int wb_init(struct cam_mgr *mgr) {
    gf_capture_id_t id = mgr->id;
    wb_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_INIT;

    return mgr->ops(mgr->disp, id, &param);
}

static int wb_deinit(struct cam_mgr *mgr) {
    gf_capture_id_t id = mgr->id;
    wb_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_DEINIT;

    return mgr->ops(mgr->disp, id, &param);
}

static int wb_enable(struct cam_mgr *mgr) {
    gf_capture_id_t id = mgr->id;
    wb_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_ENABLE;

    return mgr->ops(mgr->disp, id, &param);
}

static int wb_disable(struct cam_mgr *mgr) {
    gf_capture_id_t id = mgr->id;
    wb_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_DISABLE;

    return mgr->ops(mgr->disp, id, &param);
}

static int wb_get_caps_num(struct cam_mgr *mgr) {
    dprintk(LOG_DEBUG, mgr, "enter");

    UNUSED(mgr);
    return 1;
}

static int wb_get_caps(struct cam_mgr *mgr, struct cam_caps *caps) {
    int ret = 0;
    gf_capture_id_t id = mgr->id;
    wb_op_params_t param = {0};
    struct cam_fmt fmt = {V4L2_PIX_FMT_RGB32, 1920, 1080, 6000, 100};

    dprintk(LOG_DEBUG, mgr, "enter");
    UNUSED(mgr);

    param.op = GF_CAPTURE_OP_QUERY_CAPS;
    ret = mgr->ops(mgr->disp, id, &param);

    if (ret) {
        // TODO remove this case
        memcpy(caps->formats, &fmt, sizeof(fmt));
        dprintk(LOG_WARNING, mgr, "query_caps failed, using default %x %dx%d @ %d/%d fps",
            caps->formats[0].format, caps->formats[0].width, caps->formats[0].height, caps->formats[0].fps_num, caps->formats[0].fps_den);
    } else {
        caps->formats[0].width = param.query_caps_param.width;
        caps->formats[0].height = param.query_caps_param.height;
        caps->formats[0].format = capture_fmt_2_v4l2_fmt(param.query_caps_param.capture_fmt);
        caps->formats[0].fps_num = param.query_caps_param.refrate;
        caps->formats[0].fps_den = REFRESHRATE_DEN;
        dprintk(LOG_WARNING, mgr, "query_caps %x %dx%d @ %d/%d fps",
            caps->formats[0].format, caps->formats[0].width, caps->formats[0].height, caps->formats[0].fps_num, caps->formats[0].fps_den);
    }
    caps->caps_num = 1;

    return 0;
}

static int wb_set_format(struct cam_mgr *mgr, struct cam_fmt *fmt) {
    gf_capture_id_t id = mgr->id;
    wb_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_SET_MODE;
    param.set_mode_param.bByPass = 1;
    param.set_mode_param.srcX = fmt->width;
    param.set_mode_param.srcY = fmt->height;
    param.set_mode_param.bDoubleBuffer = 0;
    param.set_mode_param.bUpdateImme = 0;
    param.set_mode_param.mode = GF_WB_MODE_P2P;
    param.set_mode_param.outFmt = v4l2_fmt_2_capture_fmt(fmt->format);

    dprintk(LOG_WARNING, mgr, "set format %x %dx%d @ %d/%d fps", fmt->format, fmt->width, fmt->height,fmt->fps_num, fmt->fps_den);
    return mgr->ops(mgr->disp, id, &param);
}

static int wb_set_callback(struct cam_mgr *mgr, capture_cb cb) {
    gf_capture_id_t id = mgr->id;
    wb_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    param.op = GF_CAPTURE_OP_SET_CALLBACK;
    param.set_callback_param.cb = cb;
    param.set_callback_param.cb_data = mgr;

    return mgr->ops(mgr->disp, id, &param);
}

static int wb_set_buffer(struct cam_mgr *mgr, unsigned int slot, unsigned long long addr) {
    gf_capture_id_t id = mgr->id;
    wb_op_params_t param = {0};
    dprintk(LOG_DEBUG, mgr, "enter");

    UNUSED(slot);
    param.op = GF_CAPTURE_OP_SET_BUFF;
    param.set_buffer_param.fbAddr = addr;

    return mgr->ops(mgr->disp, id, &param);
}

static struct cam_ops wb_ops = {
    .init           = wb_init,
    .deinit         = wb_deinit,
    .enable         = wb_enable,
    .disable        = wb_disable,
    .get_caps_num   = wb_get_caps_num,
    .get_caps       = wb_get_caps,
    .set_format     = wb_set_format,
    .set_callback   = wb_set_callback,
    .set_buffer     = wb_set_buffer,
};

/************************************************************************/
// fake image
//
//
/************************************************************************/

static int fake = 0;   //for fake camera
module_param(fake, int, 0644);

struct cam_mgr *g_mgr = NULL;
struct task_struct *g_kthread = NULL;
struct workqueue_struct *g_videowork_queue = NULL;
struct work_struct g_video_work;

static unsigned long g_start_jiffie = 0;
static unsigned long g_frame_cnt = 0;
struct cam_fmt g_fmt;

static inline struct cam_buf *mgr_queue_find_available(struct cam_mgr *mgr);
static inline void mgr_queue_del(struct cam_mgr *mgr, struct cam_buf *buf);

static void draw_demo(void *pData, unsigned int width, unsigned int height, unsigned int new_x) {
    unsigned char *pStart = (unsigned char *) pData;
    int h, w;
    if (g_fmt.format == V4L2_PIX_FMT_YUYV) {
        memset(pStart, new_x, width * height * 2);

        dprintk(LOG_WARNING, g_mgr, "draw_demo--------new_x %d, format %s\n", new_x, "YUYV");
    } else if (g_fmt.format == V4L2_PIX_FMT_RGB32) {
        memset(pStart, new_x, width * height * 4);
        //dprintk(LOG_WARNING, g_mgr, "draw_demo--------new_x %d, format %s\n", new_x, "ARGB32");
    } else if (g_fmt.format == V4L2_PIX_FMT_NV12) {

    }
}

static void fake_thread_tick(void) {
    if (g_videowork_queue)
        queue_work(g_videowork_queue, &g_video_work);
    //else
    //printk("KMD: can't find the h8_v4l2_camera work queue\n");
}

static void fake_work(struct work_struct *work) {
    struct cam_mgr *mgr = g_mgr;
    struct cam_buf *new_buf = NULL;
    void *cpu_vaddr = NULL;
    static unsigned char new_x = 0;

    dprintk(LOG_DEBUG, mgr, "enter");
    QUEUE_LOCK(mgr);

    {
        new_buf = mgr_queue_find_available(mgr);
        WARN_ON_ONCE(!new_buf);

        new_buf->flag = CAM_BUF_USED;

        if (likely(!mgr->notify_disable && mgr->cb)) {
            mgr_queue_del(mgr, new_buf);
            cpu_vaddr = cam_buf_get_cpu_vaddr(new_buf);
            if (cpu_vaddr) {
                draw_demo(cpu_vaddr, 640, 480, new_x++);
            }
            //dprintk(LOG_WARNING, mgr, "cb_v4l2 buf %p addr %llx, cpu addr %p", new_buf, cam_buf_get_addr(new_buf), cpu_vaddr);
            mgr->cb(mgr->cb_data, new_buf, CAM_BUF_FLAG_NONE);
            mgr->notify_count++;
            new_buf->flag = CAM_BUF_AVAILABLE;
        }
    }

    QUEUE_UNLOCK(mgr);
}

static void myDelay(void) {
    unsigned long j1 = g_frame_cnt * HZ / 30 + g_start_jiffie;   // delay 33.3ms per every frame
    while (time_before(jiffies, j1)) {
        schedule();
    }
    g_frame_cnt++;
}

static int fake_thread(void *data) {
    g_frame_cnt = 0;
    g_start_jiffie = jiffies;

    set_freezable();
    for (;;) {
        fake_thread_tick();

        myDelay();

        if (kthread_should_stop())
            break;
    }

    dprintk(LOG_WARNING, g_mgr, "fake thread exit");
    return 0;
}

static int vip_start_fake(void) {
    char workqueue_name[32];
    memset(workqueue_name, 0, 32);
    sprintf(workqueue_name, "vip_camera_fake%d", 0);
    g_videowork_queue = create_singlethread_workqueue(workqueue_name);
    if (!g_videowork_queue) {
        printk("KMD: Can't create work queue.\n");
        return -EFAULT;
    }
    INIT_WORK(&g_video_work, fake_work);

    g_kthread = kthread_run(fake_thread, NULL, "vip_camera_fake");

    if (IS_ERR(g_kthread)) {
        return PTR_ERR(g_kthread);
    }

    return 0;
}

static void vip_stop_fake(void) {
    /* shutdown control thread */
    if (g_kthread) {
        kthread_stop(g_kthread);
        g_kthread = NULL;
    }

    cancel_work_sync(&g_video_work);
    memset(&g_video_work, 0, sizeof(g_video_work));

    if (g_videowork_queue) {
        flush_workqueue(g_videowork_queue);
        destroy_workqueue(g_videowork_queue);
        g_videowork_queue = NULL;
    }
}

static int vip_init_fake(struct cam_mgr *mgr) {
    dprintk(LOG_DEBUG, mgr, "enter");

    g_mgr = mgr;

    return 0;
}

static int vip_deinit_fake(struct cam_mgr *mgr) {
    dprintk(LOG_DEBUG, mgr, "enter");


    return 0;
}

static int vip_enable_fake(struct cam_mgr *mgr) {
    dprintk(LOG_WARNING, mgr, "enter");

    vip_start_fake();

    return 0;
}

static int vip_disable_fake(struct cam_mgr *mgr) {
    dprintk(LOG_WARNING, mgr, "enter");

    vip_stop_fake();

    return 0;
}

static int vip_get_caps_num_fake(struct cam_mgr *mgr) {
    dprintk(LOG_WARNING, mgr, "enter");

    return 1;
}

static int vip_get_caps_fake(struct cam_mgr *mgr, struct cam_caps *caps) {
    int ret = 0;
    dprintk(LOG_WARNING, mgr, "enter");

    caps->formats[0].width = 1920;
    caps->formats[0].height = 1080;
    caps->formats[0].format = V4L2_PIX_FMT_RGB32;
    caps->formats[0].fps_num = 3000;
    caps->formats[0].fps_den = REFRESHRATE_DEN;
    return ret;
}

static int vip_set_format_fake(struct cam_mgr *mgr, struct cam_fmt *fmt) {

    memcpy(&g_fmt, fmt, sizeof(g_fmt));
    return 0;
}

static int vip_set_callback_fake(struct cam_mgr *mgr, capture_cb cb) {
    return 0;
}

static int vip_set_buffer_fake(struct cam_mgr *mgr, unsigned int slot, unsigned long long addr) {
    return 0;
}

static struct cam_ops vip_ops_fake = {
    .init           = vip_init_fake,
    .deinit         = vip_deinit_fake,
    .enable         = vip_enable_fake,
    .disable        = vip_disable_fake,
    .get_caps_num   = vip_get_caps_num_fake,
    .get_caps       = vip_get_caps_fake,
    .set_format     = vip_set_format_fake,
    .set_callback   = vip_set_callback_fake,
    .set_buffer     = vip_set_buffer_fake,
};
/**************************************************************************/

static inline void mgr_queue_add(struct cam_mgr *mgr, struct cam_buf *buf) {
    UNUSED(mgr);
    list_add(&buf->node, &mgr->queue);
    mgr->queue_cnt++;
    dprintk(LOG_INFO, mgr, "queue_cnt %d buf %p addr %llx", mgr->queue_cnt, buf, cam_buf_get_addr(buf));
}

static inline void mgr_queue_add_tail(struct cam_mgr *mgr, struct cam_buf *buf) {
    list_add_tail(&buf->node, &mgr->queue);
    mgr->queue_cnt++;
    dprintk(LOG_INFO, mgr, "queue_cnt %d buf %p addr %llx", mgr->queue_cnt, buf, cam_buf_get_addr(buf));
}

static inline void mgr_queue_del(struct cam_mgr *mgr, struct cam_buf *buf) {
    UNUSED(mgr);
    list_del(&buf->node);
    mgr->queue_cnt--;
    dprintk(LOG_INFO, mgr, "queue_cnt %d buf %p addr %llx", mgr->queue_cnt, buf, cam_buf_get_addr(buf));
}

static inline struct cam_buf *mgr_queue_find_available(struct cam_mgr *mgr) {
    struct cam_buf *tmp = NULL, *buf = NULL, *ret = NULL;

    do {
        if (list_empty(&mgr->queue))
            break;

        list_for_each_entry_safe(buf, tmp, &mgr->queue, node) {
            if (buf->flag == CAM_BUF_AVAILABLE) {
                ret = buf;
                break;
            } else if (buf->flag == CAM_BUF_BUSY && cam_buf_is_singaled(buf)) {
                buf->flag = CAM_BUF_AVAILABLE;
                ret = buf;
                break;
            }
        }
    } while (0);

//    if(ret)
//        dprintk(LOG_INFO, mgr, "queue_cnt %d buf %p addr %llx", mgr->queue_cnt, ret, cam_buf_get_addr(ret));

    return ret;
}

/**************************************************************************/

static ssize_t mgr_proc_read(struct file *filp, char __user *buf, size_t count, loff_t *offp) {
    return 0;
}

#define PROC_BUFFER_SIZE 16

static ssize_t mgr_proc_write(struct file *filp, const char __user *buf, size_t count, loff_t *offp) {
    struct cam_mgr *mgr = PDE_DATA(file_inode(filp));
    struct cam_buf *cur = NULL, *next = NULL;

    char temp[PROC_BUFFER_SIZE];
    size_t size = count;
    int level;
    char mode = '0';

    if (count > PROC_BUFFER_SIZE)
        size = PROC_BUFFER_SIZE;

    if (copy_from_user(&temp[0], buf, size)) {
        return -EFAULT;
    }
    temp[size-1] = '\0';
//    mgr_warn(mgr, "debug info: receive msg from proc:%s count %ld", temp, count);

    if (size == 2) {
        mode = temp[0];
        if (mode == '0') {
            mgr_warn(mgr, "debug info: irq_count %lld notify_count %lld", mgr->irq_count, mgr->notify_count);
        } else if (mode == '1') {
            QUEUE_LOCK(mgr);
            mgr_warn(mgr, "debug info: queue buffer count %d", mgr->queue_cnt);
            list_for_each_entry_safe(cur, next, &mgr->queue, node) {
                mgr_warn(mgr, "debug info: buf %p addr %llx flag %s", cur, cam_buf_get_addr(cur), cam_buf_flag_2_str(cur));
            }
            QUEUE_UNLOCK(mgr);
        } else if (mode == '9') {
            mgr->notify_disable = ~mgr->notify_disable;
            mgr_warn(mgr, "debug info: toggle disable_notify %d", mgr->notify_disable);
        }
    } else if (size == 4 && temp[0] == 'l' && temp[1] == ' ') {
        if (temp[2] >= '0' && temp[2] <= '3') {
            level = temp[2] - '0';
            mgr_debuglevel = level;
        }
    } else {
        mgr_warn(mgr, "error proc cmd");
    }

    return size;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0)
const struct proc_ops mgr_proc_ops = {
    //.read   = mgr_proc_read,
    .proc_write  = mgr_proc_write,
};

#else
const struct file_operations mgr_proc_ops = {
    //.read   = mgr_proc_read,
    .write  = mgr_proc_write,
};
#endif

/************************************************************************/

static void SET_SLOT(struct cam_mgr *mgr, struct cam_buf *buf) {
    mgr->hw_buf[mgr->buf_set] = buf;
    mgr->buf_set++;
    mgr->buf_set %= HARDWARE_BUFFER_MAX;

    mgr->cam_ops->set_buffer(mgr, mgr->cur_slot, cam_buf_get_addr(buf));
    dprintk(LOG_INFO, mgr, "set_slot %d buf %p addr %llx", mgr->cur_slot, buf, cam_buf_get_addr(buf));

    mgr->cur_slot++;
    mgr->cur_slot %= mgr->slot_cnt;
}

static struct cam_buf *GET_SLOT(struct cam_mgr *mgr) {
    struct cam_buf *buf = NULL;
    buf = mgr->hw_buf[mgr->buf_get];
    mgr->hw_buf[mgr->buf_get] = NULL;
    mgr->buf_get++;
    mgr->buf_get %= HARDWARE_BUFFER_MAX;

    dprintk(LOG_INFO, mgr, "get_slot %d buf %p addr %llx", mgr->cur_slot, buf, cam_buf_get_addr(buf));
    return buf;
}


static void mgr_pause_internal(struct cam_mgr *mgr) {
    STATE_LOCK(mgr);

    do{
        if(STATE_CURRENT(mgr) == CAM_STATE_PAUSED)
            break;
        else if(STATE_CURRENT(mgr) != CAM_STATE_RUNNING){
            WARN_ON_ONCE(1);
        }

        dprintk(LOG_WARNING, mgr, "transit to pause");
        STATE_TRANSITION_START(mgr, CAM_STATE_PAUSED);

        // disable
        mgr->cam_ops->disable(mgr);
        STATE_TRANSITION_COMMIT(mgr);
    }while (0);

    dprintk(LOG_WARNING, mgr, "transit to pause complete");
    STATE_UNLOCK(mgr);
}

static void mgr_run_internal(struct cam_mgr *mgr) {
    struct cam_buf *tmp = NULL, *buf = NULL;
    unsigned int i;
    int ret;

    STATE_LOCK(mgr);
    do {
        if(STATE_CURRENT(mgr) == CAM_STATE_RUNNING)
            break;
        else if(STATE_CURRENT(mgr) != CAM_STATE_PAUSED){
            WARN_ON_ONCE(1);
        }

        dprintk(LOG_WARNING, mgr, "transit to running");
        STATE_TRANSITION_START(mgr, CAM_STATE_RUNNING);
        //FIXME. need set format again ?
        mgr_set_format(mgr, &mgr->fmt);

        // reset some field
        // refresh USED buffer, and keep it, when we change to running, we owned enough buffer
        QUEUE_LOCK(mgr);
        mgr->cur_slot = 0;
        mgr->buf_get = 0;
        mgr->buf_set = 0;
        for(i = 0; i < HARDWARE_BUFFER_MAX; i++){
            mgr->hw_buf[i] = NULL;
        }
        list_for_each_entry_safe(buf, tmp, &mgr->queue, node) {
            if (buf->flag == CAM_BUF_USED) {
                buf->flag = CAM_BUF_AVAILABLE;
            }
        }

        i = mgr->slot_cnt;
        while (i > 0) {
            buf = mgr_queue_find_available(mgr);
            if (!buf)
                continue;

            SET_SLOT(mgr, buf);
            buf->flag = CAM_BUF_USED;
            i--;
        }
        QUEUE_UNLOCK(mgr);

        ret = mgr->cam_ops->enable(mgr);

        if (ret != 0)
            STATE_TRANSITION_CANCEL(mgr);
        else
            STATE_TRANSITION_COMMIT(mgr);
    }while (0);
    dprintk(LOG_WARNING, mgr, "transit to running complete");
    STATE_UNLOCK(mgr);
}

static void capture_irq(gf_capture_id_t id, void *data, unsigned int flags) {
    struct cam_mgr *mgr = (struct cam_mgr *) data;
    struct cam_buf *done_buf = NULL, *new_buf = NULL;
    int i;
    int fmt_num = 0;
    struct cam_fmt *fmts = NULL;
    struct cam_caps caps = {0};

    dprintk(LOG_DEBUG, mgr, "enter");
    WARN_ON_ONCE(mgr->id != id);

    /**
     * TODO
     *
     * for write-back, there 3 cases when hot-plug or resolution-change happened:
     *
     * case 1 : plug-out --> plug-in --> change_pre --> change-post
     * case 2 : plug-out --> plug-in --> keep
     * case 3 : change-pre --> change-post
     *
     * for case 1 & 3, we need restart restart wb
     * for case 2, we just need disable & enable wb
     *
     * vip is different, need further check
     *
     */
    if (mgr->type == CAM_WB) {
        if (flags & GF_CAPTURE_PLUG_OUT
            || flags & GF_CAPTURE_PLUG_IN
            || flags & GF_CAPTURE_SIGNAL_OFF) {
            // TODO
            // when plug out, we should send NO_SIGNAL frame
            // currently we just pause
            dprintk(LOG_WARNING, mgr, "receive flag %d", flags);
            mgr_pause_internal(mgr);
        } else if (flags & GF_CAPTURE_SIGNAL_ON) {
            dprintk(LOG_WARNING, mgr, "receive flag GF_CAPTURE_SIGNAL_ON");
            do {
                fmt_num = mgr_get_caps_num(mgr);
                if (fmt_num <= 0) break;

                fmts = kmalloc(sizeof(struct cam_fmt) * fmt_num, GFP_KERNEL | __GFP_ZERO);
                if (!fmts){
                    dprintk(LOG_ERR, mgr, "kmalloc failed");
                    break;
                }

                caps.caps_num = fmt_num;
                caps.formats = fmts;
                if (mgr_get_caps(mgr, &caps)){
                    kfree(fmts);
                    break;
                }
                dprintk(LOG_WARNING, mgr, "get caps success");
                dprintk(LOG_WARNING, mgr, "previous format %x %dx%d @ %d/%d fps",
                    mgr->fmt.format, mgr->fmt.width, mgr->fmt.height, mgr->fmt.fps_num, mgr->fmt.fps_den);

                //CHECK current format still support or not
                for (i = 0; i < fmt_num; i++) {
                    if (fmts[i].format == mgr->fmt.format
                            && fmts[i].width == mgr->fmt.width
                            && fmts[i].height == mgr->fmt.height
                            && (fmts[i].fps_num * mgr->fmt.fps_den == fmts[i].fps_den * mgr->fmt.fps_num)){
                        dprintk(LOG_WARNING, mgr, "current format match with previous format, go on");
                        break;
                    }
                }
                kfree(fmts);

                if(i == fmt_num){
                    dprintk(LOG_WARNING, mgr, "current format not match with previous format, notify upper layer");
                    break;
                }

                // resolution keep same, we just re-running
                mgr_run_internal(mgr);
                return;
            } while (0);

            // TODO
            // if resolution changed, we need notify V4L2 and re-start
            QUEUE_LOCK(mgr);

            // we got re_acquire all buffer in slot
            /*
            for(i = 0; i < HARDWARE_BUFFER_MAX; i++){
                dprintk(LOG_WARNING, mgr, "hw_buf[%d] %p", i, mgr->hw_buf[i]);
                done_buf = mgr->hw_buf[i];
                if(done_buf){
                    done_buf->flag = CAM_BUF_UNUSED;
                    break;
                }
            }
            */

            if(mgr->cb){
                done_buf = GET_SLOT(mgr);
                done_buf->flag = CAM_BUF_UNUSED;
                dprintk(LOG_WARNING, mgr, "notify resolution change with buffer %p", done_buf);
                mgr_queue_del(mgr, done_buf);
                mgr->cb(mgr->cb_data, done_buf, CAM_BUF_FLAG_RESOLUTION_CHANGED);
            }else{
                dprintk(LOG_WARNING, mgr, "cb is null");
            }
            QUEUE_UNLOCK(mgr);

            return;
        }
    }

    STATE_LOCK(mgr);
    if (STATE_CURRENT(mgr) != CAM_STATE_RUNNING
            && STATE_CURRENT(mgr) != CAM_STATE_INIT){
        STATE_UNLOCK(mgr);
        return;
    }
    STATE_UNLOCK(mgr);

    QUEUE_LOCK(mgr);
    if (flags & GF_CAPTURE_SKIP_FRAME) {
        mgr_info(mgr, "skip frame");
    } else {
        done_buf = GET_SLOT(mgr);
        done_buf->flag = CAM_BUF_UNUSED;
        mgr->irq_count++;
    }

    // this flag should always exists
    //if(likely(flags & GF_CAPTURE_SET_BUFFER))
    {
        new_buf = mgr_queue_find_available(mgr);
        WARN_ON_ONCE(!new_buf && !done_buf);

        if (!new_buf && done_buf) {
            new_buf = done_buf;
            done_buf = NULL;
        }

        SET_SLOT(mgr, new_buf);
        new_buf->flag = CAM_BUF_USED;

        if (likely(!mgr->notify_disable && done_buf && mgr->cb)) {
            mgr_queue_del(mgr, done_buf);
            dprintk(LOG_INFO, mgr, "cb_v4l2 buf %p addr %llx", done_buf, cam_buf_get_addr(done_buf));
            mgr->cb(mgr->cb_data, done_buf, CAM_BUF_FLAG_NONE);
            mgr->notify_count++;
        } else if (done_buf) {
            done_buf->flag = CAM_BUF_AVAILABLE;
        }
    }
//    else {
//        WARN_ON_ONCE(1);
//    }
    QUEUE_UNLOCK(mgr);
}

/**************************************************************************/
//
//  below is public api
//
//
/**************************************************************************/

int mgr_init(unsigned int card_idx, unsigned int index, struct cam_mgr **pmgr) {
    struct cam_mgr *mgr;
    int ret = 0;

    if (index > GF_CAPTURE_WB4) {
        printk(KERN_ERR "out of index range or cam_mgr has been inited\n");
        return -EBUSY;
    }

    mgr = kmalloc(sizeof(struct cam_mgr), GFP_KERNEL | __GFP_ZERO);
    if (mgr == NULL) {
        printk(KERN_ERR "kmalloc failed, no enough memory\n");
        return -ENOMEM;
    }

    UNUSED(card_idx);
    mgr->id = (gf_capture_id_t) index;
    if (index < GF_CAPTURE_WB1) {
        sprintf(mgr->name, "vip%d", mgr->id+1);
        mgr->type = CAM_VIP;
        if (fake)
            mgr->cam_ops = &vip_ops_fake;
        else
            mgr->cam_ops = &vip_ops;
        mgr->slot_cnt = VIP_SLOT_BUFFER_NUM;
        mgr->enable_cnt = VIP_ENABLE_BUFFER_NUM;
    } else {
        sprintf(mgr->name, "wb%d", mgr->id+1);
        mgr->type = CAM_WB;
        mgr->cam_ops = &wb_ops;
        mgr->slot_cnt = WB_SLOT_BUFFER_NUM;
        mgr->enable_cnt = WB_ENABLE_BUFFER_NUM;
    }
    mgr->cur_slot = 0;
    mgr->buf_get = 0;
    mgr->buf_set = 0;

    INIT_LIST_HEAD(&mgr->queue);
    spin_lock_init(&mgr->queue_lock);
    mgr->queue_cnt = 0;

    STATE_INITIALIZE(mgr);

    //debug
    mgr->irq_count = 0;
    mgr->notify_count = 0;
    mgr->notify_disable = 0;

    STATE_LOCK(mgr);
    STATE_TRANSITION_START(mgr, CAM_STATE_INIT);
    do {	  
        // TODO CAPTURE INIT
	#undef __CONCAT
        #undef CONCAT
        #define __CONCAT(x,y) x##y
        #define CONCAT(x,y)     __CONCAT(x,y)

        CONCAT(gf_krnl_get_capture_interface_, DRIVER_NAME)(mgr->id, &mgr->disp, &mgr->ops);

        #undef __CONCAT
        #undef CONCAT

        if (!mgr->disp || !mgr->ops) {
            mgr_err(mgr, "get interface failed");
            ret = -ENODEV;
            break;
        }

        if (mgr->cam_ops->init(mgr)) {
            mgr_err(mgr, "init failed");
            ret = -ENODEV;
            break;
        }
        dprintk(LOG_DEBUG, mgr, "init success");
    } while (0);

    if(!ret){
        // TODO create proc entry
        mgr->proc_entry = proc_create_data(mgr->name, S_IWUSR | S_IWGRP | S_IWOTH, NULL, &mgr_proc_ops, mgr);
        if (!mgr->proc_entry) {
            mgr_err(mgr, "create proc failed, ignored");
        }
        *pmgr = mgr;
        STATE_TRANSITION_COMMIT(mgr);
        STATE_UNLOCK(mgr);
    } else{
        STATE_TRANSITION_CANCEL(mgr);
        STATE_UNLOCK(mgr);
        kfree(mgr);
    }

    return ret;
}

int mgr_deinit(struct cam_mgr *mgr) {
    if (!mgr)
        return -EINVAL;

    dprintk(LOG_DEBUG, mgr, "enter");

    STATE_LOCK(mgr);
    STATE_TRANSITION_START(mgr, CAM_STATE_DEINIT);
    if (mgr->cam_ops->deinit(mgr)) {
        mgr_err(mgr, "deinit failed");
    }

    if (mgr->proc_entry) {
        remove_proc_entry(mgr->name, NULL);
    }
    dprintk(LOG_ERR, mgr, "debug info: irq_count %lld notify_count %lld", mgr->irq_count, mgr->notify_count);

    STATE_TRANSITION_COMMIT(mgr);
    STATE_UNLOCK(mgr);
    kfree(mgr);

    return 0;
}

int mgr_get_caps_num(struct cam_mgr *mgr) {
    if (!mgr)
        return -EINVAL;

    dprintk(LOG_DEBUG, mgr, "enter");
    return mgr->cam_ops->get_caps_num(mgr);
}

int mgr_get_caps(struct cam_mgr *mgr, struct cam_caps *caps) {
    if (!mgr || !caps || caps->caps_num <= 0 || !caps->formats)
        return -EINVAL;

    dprintk(LOG_DEBUG, mgr, "enter");
    return mgr->cam_ops->get_caps(mgr, caps);
}

int mgr_set_format(struct cam_mgr *mgr, struct cam_fmt *fmt) {
    int ret;

    if (!mgr || !fmt)
        return -EINVAL;

    dprintk(LOG_DEBUG, mgr, "enter");
    ret = mgr->cam_ops->set_format(mgr, fmt);
    if (!ret)
        mgr->fmt = *fmt;
    else
        mgr_err(mgr, "set_format failed");

    return ret;
}

int mgr_set_callback(struct cam_mgr *mgr, update_buffer_func func, void *data) {
    if (!mgr || !func)
        return -EINVAL;

    dprintk(LOG_DEBUG, mgr, "enter");

    // TODO set callback, need check ret?
    mgr->cam_ops->set_callback(mgr, capture_irq);

    // save data
    mgr->cb_data = data;
    mgr->cb = func;

    return 0;
}

int mgr_queue_buffer(struct cam_mgr *mgr, struct cam_buf *buffer) {
    if (!mgr || !buffer)
        return -EINVAL;

    STATE_LOCK(mgr);
    if(STATE_CURRENT(mgr) != CAM_STATE_INIT
            && STATE_CURRENT(mgr) != CAM_STATE_RUNNING
            && STATE_CURRENT(mgr) != CAM_STATE_PAUSED){
        STATE_UNLOCK(mgr);
        return -EINVAL;
    }

    dprintk(LOG_DEBUG, mgr, "enter");

    if (buffer->flag != CAM_BUF_UNUSED) {
        // warning!
    }

    if (cam_buf_is_singaled(buffer))
        buffer->flag = CAM_BUF_AVAILABLE;
    else
        buffer->flag = CAM_BUF_BUSY;

    QUEUE_LOCK(mgr);
    mgr_queue_add_tail(mgr, buffer);
    QUEUE_UNLOCK(mgr);
    STATE_UNLOCK(mgr);

    return 0;
}

int mgr_enable(struct cam_mgr *mgr) {
    struct cam_buf *buf;
    unsigned int i;
    int ret;
    //struct cam_fmt tmp = {V4L2_PIX_FMT_RGB32, 1920, 1080, 6000, 100};

    if (!mgr)
        return -EINVAL;

    // TODO check queue_cnt, must not less than enable_cnt
    if (mgr->queue_cnt < mgr->enable_cnt)
        return -ENOMEM;

    STATE_LOCK(mgr);
    if(STATE_CURRENT(mgr) != CAM_STATE_INIT){
        STATE_UNLOCK(mgr);
        return -1;
    }

    dprintk(LOG_DEBUG, mgr, "enter");

    STATE_TRANSITION_START(mgr, CAM_STATE_RUNNING);
    //FIXME. set format, need remove later
    //if(mgr->fmt.width == 0 || mgr->fmt.height == 0)
    //    mgr_set_format(mgr, &tmp);

    i = mgr->slot_cnt;
    QUEUE_LOCK(mgr);
    while (i > 0) {
        buf = mgr_queue_find_available(mgr);
        if (!buf)
            continue;

        SET_SLOT(mgr, buf);
        buf->flag = CAM_BUF_USED;
        i--;
    }
    QUEUE_UNLOCK(mgr);

    ret = mgr->cam_ops->enable(mgr);

    if (ret != 0)
        STATE_TRANSITION_CANCEL(mgr);
    else
        STATE_TRANSITION_COMMIT(mgr);

    STATE_UNLOCK(mgr);
    return ret;
}

int mgr_disable(struct cam_mgr *mgr) {
    struct cam_buf *buf;
    if (!mgr)
        return -EINVAL;

    STATE_LOCK(mgr);
    if(STATE_CURRENT(mgr) != CAM_STATE_RUNNING
            && STATE_CURRENT(mgr) != CAM_STATE_PAUSED){
        STATE_UNLOCK(mgr);
        return -EIO;
    }

    dprintk(LOG_DEBUG, mgr, "enter");

    if(STATE_CURRENT(mgr) == CAM_STATE_RUNNING){
        mgr->cam_ops->disable(mgr);
    }
    STATE_TRANSITION_START(mgr, CAM_STATE_STOPPED);

    QUEUE_LOCK(mgr);
    while (!list_empty(&mgr->queue)) {
        buf = list_first_entry(&mgr->queue, struct cam_buf, node);
        buf->flag = CAM_BUF_UNUSED;
        mgr_queue_del(mgr, buf);
        mgr->cb(mgr->cb_data, buf, CAM_BUF_FLAG_NONE);
    }
    QUEUE_UNLOCK(mgr);
    dprintk(LOG_WARNING, mgr, "send all buffers back to v4l2 framework");

    STATE_TRANSITION_COMMIT(mgr);
    STATE_UNLOCK(mgr);
    return 0;
}

